/* metabol: merge two .mps files, invoke lp_solve and process its result */
/* L.Rubanov, 2006 */
/* 22.04.06  1.4  Fix typos in some messages */
/* 02.04.06  1.3  Already re-format result, even in -r case. Debug confition!
						Change interim filenames to <model>.in and <task>.out */
/* 31.03.06  1.2  lp_solve.c text embedded & liblpk library attached
						NB: static 'stream' in the library must be external !
                  Rename NAMELEN*->RNAMLEN*. Post-processing of result file */
/* 29.03.06  1.1  Bug fixed, some enhancements. Unclear memory problem exists */
/* 27.03.06  1.0	Base version */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "lpkit.h"
#include "lpglob.h"
#include "patchlevel.h"

#define	VERSION	"1.4"
#define	MAXABBR	1000			/* max number of metabolites */
#define	ADDABBR	10				/* max number of additional metabolites */
#define	ABBRLEN	8				/* max length of metabolite name */
#define	ABBRLENS	"8"			/* the same as string */
#define	MAXREACT 2000			/* max number of reactions */
#define	ADDREACT	100			/* max number of additional reactions */
#define	RNAMLEN	8				/* max length of reaction name */
#define	RNAMLENS "8"			/* the same as string */
#define	MAXBOUND	3000			/* max number of bounds */
#define  ADDBOUND	200			/* max number of additional bounds */
#define	BUFLEN	64000			/* length of line buffer */
#define	TRIVINF	1.0e+20		/* trivial infinity */

typedef struct {
	char	abbr[ABBRLEN+1];
   char	modifier;
}	META;

typedef struct {
	char	name[RNAMLEN+1];
   int	lines;
   char	*mpslines;
}	REACTION;

typedef struct {
	char	name[RNAMLEN+1];
   char	type[3];		/* LO, UP, FX ... */
   double value;
}  BOUNDS;

/* print_help, print_cpu_times, lp_solve (former main) are taken from lp_solve */
void print_help(void);
void print_cpu_times(const char *info);
int lp_solve( int argc, char **argv, FILE *infile, FILE *outfile );
extern FILE	*stream;							/* LR: output file in debug.c */

/*----------------------------------------------------------------------------*/
int main( int argc, char** argv ) {
	META		*meta, *addmeta;
   REACTION	*react, *addreact;
   BOUNDS	*bound, *addbound;
	FILE		*model=NULL, *task=NULL, *out=NULL, *mps=NULL, *res=NULL;
   int		metanum, addmetanum, reactnum, addreactnum, boundnum, addboundnum;
   int 		i, l, rhs=0;
   int		na, nummod=0, reduce=1, rmlines=0, rtlines=0, tlines=0;
   char		*wmode="wt", *buf, tmp[256], abbr[ABBRLEN+1], name[RNAMLEN+1];
   double	d, f, t;
   printf("\n(C) 2006, Laboratory of Mathematical Methods and Models in Bioinformatics\n"
          "Institute for Information Transmission Problems, Russian Academy of Sciences\n"
   		 "Program for Metabolic Flux Balance Analysis   v.%s          L.Rubanov, 2006\n\n",
          VERSION);
	if( argc<4 ) {
   	printf(
      	"Usage: metabol [options] model_file task_file out_file\n\n"
         "-a          append to out_file (rewrite by default)\n"
         "-r          do not reduce output results (delete zeros by default)\n"
         "Other options are passed directly into lp_solve library.\n\n" );
      print_help( );
   	printf(
         "model_file: file in MPS format, containing general model of metabolism\n"
         "task_file:  file in ~MPS format, containing additions and/or changes\n"
         "out_file:   output text file with solution of the task\n\n" );
      printf( "Thank you for using the program.\n\n" );
      return 0;
   }
   for( na=1; na<argc; na++ ) {		/* parse options and filenames */
      if( argv[na][0]=='-' ) {
      	if( argv[na][1]=='r' || argv[na][1]=='R' )  reduce=0;
         else if( argv[na][1]=='a' || argv[na][1]=='A' ) wmode[0]='a';
         else continue;							/* skip other options */
         for( i=na+1; i<argc; i++ )			/* remove my options */
         	argv[i-1]=argv[i];
         argc--; na--;
      }
      else if( model==NULL ) {
   		if( (model=fopen( argv[na], "rt" ))==NULL )
         	{ printf( "Can't open file '%s'\n", argv[na] ); return -1; }
         printf( "Model: %s", argv[na] );
         sprintf( tmp, "%s.in", argv[na] );
   		if( (mps=fopen( tmp, "w+t" ))==NULL )
         	{ printf( "Can't open file '%s'\n", tmp ); return -1; }
         for( i=na+1; i<argc; i++ )			/* remove model name from arguments */
         	argv[i-1]=argv[i];
         argc--; na--;
      }
      else if( task==NULL ) {
   		if( (task=fopen( argv[na], "rt" ))==NULL )
         	{ printf( "Can't open file '%s'\n", argv[na] ); return -1; }
         printf( "  Task: %s\n", argv[na] );
         sprintf( tmp, "%s.out", argv[na] );		/* open interim result file */
         if( (res=fopen( tmp, "w+t" ))==NULL )
            { printf( "Can't open file '%s'\n", tmp ); return -1; }
         for( i=na+1; i<argc; i++ )			/* remove task name from arguments */
         	argv[i-1]=argv[i];
         argc--; na--;
      }
      else if( out==NULL ) {
      	if( (out=fopen( argv[na], wmode ))==NULL )
         	{ printf( "Can't open file '%s'\n", argv[na] ); return -1; }
         for( i=na+1; i<argc; i++ )			/* remove result filename from args */
         	argv[i-1]=argv[i];
         argc--; na--;
      }
   }
   if( !model || !task || !out )
   	{ printf( "Not enough files specified (need 3)\n" ); return -1; }
   buf=malloc( BUFLEN );
   meta=(META*)calloc( MAXABBR, sizeof(META) );
   addmeta=(META*)calloc( ADDABBR, sizeof(META) );
   react=(REACTION*)calloc( MAXREACT, sizeof(REACTION) );
   addreact=(REACTION*)calloc( ADDREACT, sizeof(REACTION) );
   bound=(BOUNDS*)calloc( MAXBOUND, sizeof(BOUNDS) );
   addbound=(BOUNDS*)calloc( ADDBOUND, sizeof(BOUNDS) );
   if( !buf || !meta || !addmeta || !react || !addreact || !bound || !addbound )
   	{ printf( "Can't allocate some structure(s)\n" ); return -1; }

   for( addmetanum=0; addmetanum<ADDABBR; ) {		/* read adds metabolites */
   	fgets( buf, BUFLEN, task ); rtlines++;
      if( feof(task) || ferror(task) )
      	{ printf( "Error in ROWS section of task in line %d\n", rtlines ); return -1; }
      if( buf[0]=='*' || strncmp( buf, "ROWS", 4 )==0 ) continue;
      if( strncmp( buf, "COLUMNS", 7 )==0 ) break;
      addmeta[addmetanum].modifier = (char)(*buf=='-'? '-': *(buf+1));
      sscanf( buf+4, "%"ABBRLENS"s", addmeta[addmetanum++].abbr );
   }
   for( metanum=0; metanum<MAXABBR; ) {		/* read metabolites from model */
   	fgets( buf, BUFLEN, model ); rmlines++;
      if( feof(model) || ferror(model) )
      	{ printf( "Error in ROWS section of model in line %d\n", rmlines ); return -1; }
      if( buf[0]=='*' || strncmp( buf, "NAME", 4)==0 ||
      	 strncmp( buf, "ROWS", 4 )==0 ) {
      	fputs( buf, mps ); tlines++;
         continue;
      }
      if( strncmp( buf, "COLUMNS", 7 )==0 ) break;
      sscanf( buf+4, "%"ABBRLENS"s", tmp );	/* read metabolite abbr */
      for( i=0; i<addmetanum; i++ )			/* chect if metabolite is deleted */
      	if( addmeta[i].modifier=='-' && strcmp( addmeta[i].abbr, tmp )==0 )
         	break;
      if( i<addmetanum ) { nummod++; continue; }	/* skip it */
      fputs( buf, mps ); tlines++;
      meta[metanum].modifier=buf[1];
      strcpy( meta[metanum++].abbr, tmp );
   }
   if( metanum>=MAXABBR )
   	{ printf( "Too many metabolites in the model\n" ); return -1; }
   printf( "%d metabolites read from the model\n", metanum+nummod );
   printf( "%d metabolites deleted in this task\n", nummod );
   nummod=0;
   for( i=0; i<addmetanum; i++ ) {		/* write additional metabolites */
   	if( addmeta[i].modifier=='-' ) continue;
      fprintf( mps, " %c  %s\n", addmeta[i].modifier, addmeta[i].abbr );
      tlines++;
      meta[metanum].modifier=addmeta[i].modifier;
      strcpy( meta[metanum++].abbr, addmeta[i].abbr );
   	if( metanum>=MAXABBR )
      	{ printf( "Too many metabolites in the task\n" ); return -1; }
      nummod++;
   }
   printf( "%d metabolites added in this task\n", nummod );
   printf( "%d metabolites written into MPS file\n", metanum );
   memset( tmp, '*', 56 ); tmp[56]=0;
   fprintf( mps, "%s\n", tmp ); tlines++;
   fputs( buf, mps ); tlines++;						/* write COLUMNS line */
   fflush( mps );
   nummod=0;
   for( addreactnum=0; addreactnum<ADDREACT; ) {	/* read adds reactions */
   	fgets( buf, BUFLEN, task ); rtlines++;
      if( feof(task) || ferror(task) )
      	{ printf( "Error in COLUMNS section of task in line %d\n", rtlines ); return -1; }
      if( buf[0]=='*' ) continue;
      if( strncmp( buf, "RHS", 3 )==0 ) { rhs=1; break; }
      if( strncmp( buf, "BOUNDS", 6 )==0 ) break;
      sscanf( buf+4, "%"RNAMLENS"s", tmp );
      if( buf[0]=='-' ) {			/* store reaction to be deleted */
       	strcpy( addreact[addreactnum].name, tmp );
      	addreact[addreactnum].mpslines=NULL;		/* deletion flag */
         addreact[addreactnum++].lines=0;
      	nummod++; continue;
      }
		sscanf( buf+14, "%"ABBRLENS"s", abbr );  /* check abbr of the reaction */
      for( i=0; i<metanum; i++ )
      	if( strcmp( meta[i].abbr, abbr )==0 )
         	break;
      if( i>=metanum ) {
      	printf( "Unknown metabolite '%s' in task line %d\n", abbr, rtlines );
         return -1;
      }
      if( strlen( buf )>40 ) {
         sscanf( buf+39, "%"ABBRLENS"s", abbr );
         for( i=0; i<metanum; i++ )
            if( strcmp( meta[i].abbr, abbr )==0 )
               break;
         if( i>=metanum ) {
            printf( "Unknown metabolite '%s' in task line %d\n", abbr, rtlines );
            return -1;
         }
      }
      if( addreactnum==0 || strcmp( addreact[addreactnum-1].name, tmp ) ) { /* new */
       	strcpy( addreact[addreactnum].name, tmp );
      	l=strlen( buf );
      	if( (addreact[addreactnum].mpslines=malloc( l+1 ))==NULL )
         	{ printf( "Cannot allocate memory for reaction\n" ); return -1; }
         strcpy( addreact[addreactnum].mpslines, buf );
         addreact[addreactnum++].lines=1;
      }
      else {	/* add line to previous reaction */
      	l=strlen( addreact[addreactnum-1].mpslines ) + strlen( buf );
      	if( (addreact[addreactnum-1].mpslines=
         	  realloc( addreact[addreactnum-1].mpslines, l+1 ))==NULL )
         	{ printf( "Cannot re-allocate memory for reaction\n" ); return -1; }
         strcat( addreact[addreactnum-1].mpslines, buf );
         addreact[addreactnum-1].lines++;
      }
   }
   name[0]=0;
   for( reactnum=0; reactnum<MAXREACT; ) {		/* read reactions from model */
   	fgets( buf, BUFLEN, model ); rmlines++;
      if( feof(model) || ferror(model) )
      	{ printf( "Error in COLUMNS section of model in line %d\n", rmlines ); return -1; }
      if( buf[0]=='*' )
      	{ fputs( buf, mps ); tlines++; continue; }
      if( strncmp( buf, "RHS", 3 )==0 || strncmp( buf, "BOUNDS", 6 )==0 )
      	break;
      sscanf( buf+4, "%"RNAMLENS"s", tmp );
      if( strcmp( tmp, name ) ) {
         strcpy( name, tmp );
         for( i=0; i<addreactnum; i++ )		/* check if reaction is deleted */
            if( addreact[i].mpslines==NULL && strcmp( addreact[i].name, tmp )==0 )
               break;
         if( i<addreactnum ) { nummod++; continue; }	/* skip it */
         react[reactnum].lines=1; react[reactnum].mpslines=NULL;
      	strcpy( react[reactnum++].name, tmp );
      }
      else react[reactnum-1].lines++;
      fputs( buf, mps ); tlines++;
   }
   if( reactnum>=MAXREACT )
   	{ printf( "Too many reactions in the model\n" ); return -1; }
   printf( "%d reactions read from the model\n", reactnum+nummod );
   printf( "%d reactions deleted in this task\n", nummod );
   nummod=0;
   for( i=0; i<addreactnum; i++ ) {		/* write additional reactions */
   	if( addreact[i].mpslines==NULL ) continue;
      fprintf( mps, "%s", addreact[i].mpslines );
      tlines+=addreact[i].lines;
      free( addreact[i].mpslines ); addreact[i].mpslines=NULL;
      strcpy( react[reactnum].name, addreact[i].name );
      react[reactnum].lines=addreact[i].lines;
      react[reactnum++].mpslines=NULL;
      if( reactnum>=MAXREACT )
         { printf( "Too many reactions in the task\n" ); return -1; }
      nummod++;
   }
   printf( "%d reactions added in this task\n", nummod );
   printf( "%d reactions written into MPS file\n", reactnum );
   memset( tmp, '*', 56 ); tmp[56]=0;
   fprintf( mps, "%s\n", tmp ); tlines++;
   fputs( buf, mps ); tlines++;						/* write RHS/BOUNDS line */
   fflush( mps );
   if( rhs && strncmp( buf, "RHS", 3 )==0 ) {	/* copy RHS section of task */
   	while( 1 ) {
         fgets( buf, BUFLEN, task ); rtlines++;
         if( feof(task) || ferror(task) )
      		{ printf( "Error in RHS section of task in line %d\n", rtlines ); return -1; }
         if( strncmp( buf, "BOUNDS", 6 )==0 ) break;
         fputs( buf, mps ); tlines++;
      }
   	fgets( buf, BUFLEN, model ); rmlines++;
      if( feof(model) || ferror(model) || strncmp( buf, "BOUNDS", 6 )!=0 )
      	{ printf( "Error in RHS section of model in line %d\n", rmlines ); return -1; }
      fputs( buf, mps ); tlines++;
   }
   for( addboundnum=0; addboundnum<ADDBOUND; ) {	/* read adds bounds */
   	fgets( buf, BUFLEN, task ); rtlines++;
      if( feof(task) || ferror(task) )
      	{ printf( "Error in BOUNDS section of task in line %d\n", rtlines ); return -1; }
      if( buf[0]=='*' ) continue;
      if( strncmp( buf, "ENDATA", 6 )==0 ) break;
      sscanf( buf+1, "%2s%*s %"RNAMLENS"s%lf", addbound[addboundnum].type,
      	addbound[addboundnum].name, &addbound[addboundnum].value );
		for( i=0; i<reactnum; i++ ) 				/* check if reaction name exists */
      	if( strcmp( react[i].name, addbound[addboundnum].name )==0 )
         	break;
      if( i>=reactnum ) {
      	printf( "Unknown reaction name '%s' in task line %d\n",
         	addbound[addboundnum].name, rtlines );
         return -1;
      }
      addboundnum++;
   }
   nummod=0;
   for( boundnum=0; boundnum<MAXBOUND; ) {	/* read bounds from model */
   	fgets( buf, BUFLEN, model ); rmlines++;
      if( feof(model) || ferror(model) )
      	{ printf( "Error in BOUNDS section of model in line %d\n", rmlines ); return -1; }
      if( buf[0]!='*' ) {
         if( strncmp( buf, "ENDATA", 6 )==0 ) break;
         sscanf( buf+1, "%2s%*s %"RNAMLENS"s%lf", bound[boundnum].type,
            bound[boundnum].name, &bound[boundnum].value );
         for( i=0; i<addboundnum; i++ )		/* check if bound is changed */
            if( strcmp( addbound[i].name, bound[boundnum].name )==0 &&
                strcmp( addbound[i].type, bound[boundnum].type )==0 ) break;
         if( i<addboundnum ) {					/* skip it */
            bound[boundnum].name[0]=0; bound[boundnum].type[0]=0;
            nummod++; continue;
         }
         boundnum++;
      }
      fputs( buf, mps ); tlines++;
   }
   if( boundnum>=MAXBOUND )
   	{ printf( "Too many bounds in the model\n" ); return -1; }
   printf( "%d bounds read from the model\n", boundnum+nummod );
   printf( "%d bounds deleted in this task\n", nummod );
   memset( tmp, '*', 56 ); tmp[56]=0;
   fprintf( mps, "%s\n", tmp ); tlines++;
   nummod=0;
   for( i=0; i<addboundnum; i++ ) {			/* write additional bounds */
   	fprintf( mps, " %2s BND       %s%*g\n", addbound[i].type,
      	addbound[i].name, 22-strlen( addbound[i].name ), addbound[i].value );
      tlines++;
      strcpy( bound[boundnum].name, addbound[i].name );
      strcpy( bound[boundnum].type, addbound[i].type );
      bound[boundnum++].value=addbound[i].value;
   	if( boundnum>=MAXBOUND )
      	{ printf( "Too many bounds in the task\n" ); return -1; }
      nummod++;
   }
   printf( "%d bounds added in this task\n", nummod );
   printf( "%d bounds written into MPS file\n", boundnum );
   memset( tmp, '*', 56 ); tmp[56]=0;
   fprintf( mps, "%s\n", tmp ); tlines++;
   fputs( buf, mps ); tlines++;						/* write ENDATA line */
   fflush( mps );
   printf( "Total %d lines read from the model file\n", rmlines );
   printf( "Total %d lines read from the task file\n", rtlines );
   printf( "Total %d lines written into MPS file\n\n", tlines );
   free( addmeta );	free( meta );
   free( addreact );
   free( addbound );	free( bound );
   fclose( model );	fclose( task );
   rewind( mps );
/*---------------------------------------------------------------------------*/
   lp_solve( argc, argv, mps, res );		/* Call optimization library */
/*---------------------------------------------------------------------------*/
   fclose( mps );
   rewind( res );					/* convert .res contents into result file */
   while( fgets( buf, BUFLEN, res ) ) {
      if( strncmp( buf, "Actual values of the variables:", 31 ) )
         fputs( buf, out );
      else break;
   }
   fflush( out );
   if( feof(res) ) goto Fin;
   if( reduce )
   	{ buf[30]=0; strcat( buf, " (nonzero only):\n" ); }
   fputs( buf, out );
   fflush( out );
   while( fgets( buf, BUFLEN, res ) ) {
      if( strncmp( buf, "Actual values of the constraints:", 33 )==0 )
         break;
      l=sscanf( buf, "%s%lf", tmp, &d );
      if( reduce && l==2 && d==0 )
         continue;						/* skip zero variables */
      if( strncmp( tmp, "__AntiBodyOf", 12 )==0 ) {
         sprintf( tmp+6, "__%s", react[i-1].name );
      }
      if( l==2 )
         sprintf( buf, "%-16s %- 12.7g\n", tmp, d );
      fputs( buf, out );
   }
   fflush( out );
   if( feof(res) ) goto Fin;
   if( reduce )
   	{ buf[32]=0; strcat( buf, " (nonzero only if any):\n" ); }
   fputs( buf, out );
   fflush( out );
   while( fgets( buf, BUFLEN, res ) ) {
      if( strncmp( buf, "Dual values with from - till limits:", 36 )==0 )
         break;
      l=sscanf( buf, "%s%lf", tmp, &d );
      if( reduce && l==2 && d==0 )
         continue;						/* skip zero constraints */
      if( l==2 )
         sprintf( buf, "%-16s %- 12.7g\n", tmp, d );
      fputs( buf, out );
   }
   fflush( out );
   if( feof(res) ) goto Fin;
   if( reduce )
   	{ buf[35]=0; strcat( buf, " (nontrivial only):\n" ); }
   fputs( buf, out );
   fflush( out );
   while( fgets( buf, BUFLEN, res ) ) {
      l = sscanf( buf, "%s%lf%lf%lf", tmp, &d, &f, &t );
      if( reduce && l==4 && d==0. && f<-TRIVINF && t>TRIVINF )
         continue;						/* skip trivial dual values */
      if( strncmp( tmp, "__AntiBodyOf", 12 )==0 ) {
         sscanf( tmp+12, "(%d)", &i );
         sprintf( tmp+6, "__%s", react[i-1].name );
      }
      if( l==4 )
         sprintf( buf, "%-16s %- 12.7g %- 12.7g %- 12.7g\n", tmp, d, f, t );
      fputs( buf, out );
   }
Fin:
	fclose( out );
	printf( "\nThank you for using the program.\n\n" );
   return 0;
}

/*-----------------------------------------------------------------------------*/
void print_help(void) {		/* slightly modified by LR */
	printf("Some useful options of lp_solve version %d.%d.%d.%d are:\n", MAJORVERSION, MINORVERSION, RELEASE, BUILD);
/*	printf("-h          prints this message\n"); */
	printf("-v<level>   verbose mode, gives flow through the program:\n");
	printf("              0=CRITICALSTOP 1=CRITICAL 2=SEVERE 3=IMPORTANT (default)\n");
	printf("              4=NORMAL 5=DETAILED 6=FULL\n");
	printf("-d          debug mode, print all intermediate results and B&B decisions\n");
	printf("-p          print the values of the dual variables\n");
	printf("-b<bound>   specify a lower bound for the objective function to the program.\n"
          "              If close enough, may speed up the calculations.\n");
	printf("-i          print all intermediate valid solutions. Can give you useful\n"
          "              solutions even if the total run time is too long\n");
/*	printf("-e<number>  specifies the epsilon which is used to determine whether a floating\n"
          "              point number is in fact an integer. Should be < 0.5\n"); */
/*	printf("-c          during branch-and-bound, take the ceiling branch first\n");
	printf("-ca         during branch-and-bound, the algorithm chooses branch\n");
	printf("-B<rule>    specify branch-and-bound rule\n");
	printf("             -B0: Select Lowest indexed non-integer column (default)\n");
	printf("             -B1: Select Random non-integer column\n");
	printf("             -B2: Select Largest deviation from an integer value\n");
	printf("             -B3: Select Best ???\n");
	printf("             -B4: Select Median value deviation from an integer value\n");
	printf("             -B5: Select Greedy ???\n"); */
	printf("-s<mode>    use automatic problem scaling\n");
	printf("             -s0: Numerical range-based scaling (-s is an equivalent)\n");
	printf("             -s1: Geometric scaling\n");
	printf("             -s2: Curtis-reid scaling\n");
	printf("-sp         also do power scaling. This option must come after -s<mode> option\n");
	printf("-sl         also do Lagrange scaling. Must come after -s<mode> option\n");
	printf("-si         also do Integer scaling. Must come after -s<mode> option\n");
	printf("-I          print info after reinverting\n");
	printf("-t          trace pivot selection\n");
/*	printf("-lp         read from LP file (default)\n");
	printf("-mps        read from MPS file\n"); */
	printf("-degen      use perturbations to reduce degeneracy, can increase instability\n");
	printf("-trej<Trej> set minimum pivot value\n");
/*	printf("-parse_only parse input file but do not calculate (i.e. check)\n"); */
	printf("-presolve   presolve problem before start optimizing\n");
/*	printf("-improve<level>  iterative improvement level:\n");
	printf("              0=none (default) 1=FTRAN only 2=BTRAN only 3=FTRAN+BTRAN\n"); */
	printf("-time       print CPU time to parse input and to calculate result\n");
/*	printf("-min        minimize the lp problem (overrules setting in file)\n"); */
	printf("-max        maximize the lp problem (minimize by default)\n");
	printf("-S<detail>  print solution. If detail ommited, then -S2 is used.\n");
	printf("              0: none, 1: objective value, 2: +variables (default)\n");
	printf("              3: +constraints, 4: +duals, 5: +lp model, 6: +lp scales\n\n");
}

void print_cpu_times(const char *info) {
	static clock_t last_time = 0;
	clock_t new_time;
	new_time = clock();
	fprintf(stderr, "CPU Time for %s: %gs (%gs total since program start)\n",
		info, (new_time - last_time) / (double) CLOCKS_PER_SEC,
		new_time / (double) CLOCKS_PER_SEC);
	last_time = new_time;
}

int lp_solve( int argc, char *argv[], FILE *infile, FILE *outfile ) {
	lprec	*lp;
/*	char	*filen; 				LR: not needed */
	int	i;
	short	verbose=IMPORTANT /* CRITICALSTOP */;
	MYBOOL debug=FALSE, print_sol=FALSE, PRINT_DUALS=FALSE, floor_first=TRUE;
	short scaling=0;
	short print_at_invert=FALSE;
	MYBOOL tracing=FALSE;
	short mps=TRUE;							/* LR: replace FALSE */
	MYBOOL anti_degen=FALSE;
	short print_timing=FALSE;
	short parse_only=FALSE;
	MYBOOL do_presolve=FALSE;
	short	objective=0;
	short	PRINT_SOLUTION=2;
	MYBOOL improve=IMPROVE_NONE, bb_rule=FIRST_SELECT, scalemode=MMSCALING;
	int result;
	REAL obj_bound = (REAL)DEF_INFINITE;
	REAL epsilon = (REAL)DEF_EPSILON;
	REAL epspivot = (REAL)DEF_EPSPIVOT;
	FILE *fpin = infile;						/* LR: replace 'stdin' */

	for(i = 1; i < argc; i++) {
		if(strncmp(argv[i], "-v", 2) == 0) {
      	if (argv[i][2]) verbose = (short) atoi(argv[i] + 2);
		else
      	verbose = NORMAL;
      }
      else if(strcmp(argv[i], "-d") == 0)
         debug = TRUE;
      else if(strcmp(argv[i], "-i") == 0)
         print_sol = TRUE;
      else if(strcmp(argv[i], "-c") == 0)
         floor_first = FALSE;
      else if(strcmp(argv[i], "-ca") == 0)
         floor_first = AUTOMATIC;
      else if(strncmp(argv[i], "-B", 2) == 0) {
         if (argv[i][2])	bb_rule = (MYBOOL) atoi(argv[i] + 2);
         else					bb_rule = FIRST_SELECT;
      }
      else if(strcmp(argv[i], "-b") == 0)
         obj_bound = atof(argv[++i]);
      else if(strcmp(argv[i], "-e") == 0) {
         epsilon = atof(argv[++i]);
         if((epsilon <= 0.0) || (epsilon >= 0.5)) {
            fprintf(stderr, "Invalid epsilon %g; 0 < epsilon < 0.5\n",
            	(double)epsilon);
            exit(EXIT_FAILURE);
         }
      }
      else if(strcmp(argv[i], "-p") == 0)
         PRINT_DUALS = TRUE;
      else if(strcmp(argv[i], "-h") == 0)
         { print_help( ); exit(EXIT_SUCCESS); }
      else if(strcmp(argv[i], "-sp") == 0)
         scalemode = (MYBOOL) (scalemode | POWERSCALE);
      else if(strcmp(argv[i], "-sl") == 0)
         scalemode = (MYBOOL) (scalemode | LAGRANGESCALE);
      else if(strcmp(argv[i], "-s2") == 0)
         scaling = 2;
      else if(strcmp(argv[i], "-si") == 0)
         scalemode = (MYBOOL) (scalemode | INTEGERSCALE);
      else if(strncmp(argv[i], "-s", 2) == 0) {
         scaling = 1;
         if (argv[i][2])	scalemode = (MYBOOL) atoi(argv[i] + 2);
         else					scalemode = MMSCALING;
      }
      else if(strcmp(argv[i], "-I") == 0)
         print_at_invert = TRUE;
      else if(strcmp(argv[i], "-t") == 0)
         tracing = TRUE;
      else if(strncmp(argv[i], "-S", 2) == 0) {
         if (argv[i][2])
            PRINT_SOLUTION = (short) atoi(argv[i] + 2);
         else
            PRINT_SOLUTION = 2;
      }
      else if(strncmp(argv[i], "-improve", 8) == 0) {
         if (argv[i][8])
            improve = (MYBOOL) atoi(argv[i] + 8);
         else
            improve = 0;
      }
/*		else if(strcmp(argv[i],"-mps") == 0)    LR: always use -mps
         mps = TRUE;
      else if(strcmp(argv[i],"-lp") == 0)
         mps = FALSE;                           */
      else if(strcmp(argv[i],"-degen") == 0)
         anti_degen = TRUE;
      else if(strcmp(argv[i],"-time") == 0) {
         if(clock() == -1)
            fprintf(stderr, "CPU times not available on this machine\n");
         else
            print_timing = TRUE;
      }
      else if(strcmp(argv[i],"-trej") == 0)
         epspivot = atof(argv[++i]);
      else if(strcmp(argv[i],"-parse_only") == 0)
         parse_only = TRUE;  /* only useful for parser software development */
      else if(strcmp(argv[i],"-presolve") == 0)
         do_presolve = TRUE;
      else if(strcmp(argv[i],"-min") == 0)
         objective = -1;
      else if(strcmp(argv[i],"-max") == 0)
         objective =  1;
/*		else if(fpin == stdin) {			LR: stream pointers are passed
         filen = argv[i];
         if(*filen == '<')		filen++;
         if((fpin = fopen(filen, "r")) == NULL) {
            fprintf(stderr,"Error, Unable to open input file '%s'\n", argv[i]);
            print_help( ); exit(EXIT_FAILURE);
         }
      }
      else {
         filen = argv[i];
         if(*filen != '>') {
            fprintf(stderr, "Error, Unrecognized command line argument '%s'\n", argv[i]);
            print_help( ); exit(EXIT_FAILURE);
         }
      }	*/
      else {						/* LR: added for proper diagnostics */
      	fprintf(stderr, "Error, Unrecognized command line argument '%s'\n", argv[i]);
         print_help( ); exit(EXIT_FAILURE);
      }
	}
   stream = outfile; 			/* LR: redirect output to result file */
	if(mps)
   	lp = read_mps(fpin, verbose);
  	else /* standard lp_solve syntax expected */
   	lp = read_lp_file(fpin, verbose, "lp" );
/*	if(fpin != stdin)
   	fclose(fpin);					LR: don't close input file */
	if(print_timing)
   	print_cpu_times("parsing input");
	if(parse_only)		exit(0);
  	if(lp == NULL)
  		{ fprintf(stderr, "Unable to read model.\n"); exit(EXIT_FAILURE); }
  	if(objective != 0) {
   	if(objective == 1)	set_maxim(lp);
    	else						set_minim(lp);
	}
  	if(PRINT_SOLUTION >= 5)
    	print_lp(lp);
  	set_print_sol(lp, print_sol);
  	set_epsilon(lp, epsilon);
  	set_epspivot(lp, epspivot);
  	set_print_duals(lp, PRINT_DUALS);
  	set_debug(lp, debug);
  	set_floor_first(lp, floor_first);
  	set_print_at_invert(lp, print_at_invert);
  	set_trace(lp, tracing);
  	if(obj_bound != DEF_INFINITE)
    	set_obj_bound(lp, obj_bound);
  	set_anti_degen(lp, anti_degen);
  	set_do_presolve(lp, do_presolve);
  	set_improve(lp, improve);
  	set_scalemode(lp, scalemode);
  	set_bb_rule(lp, bb_rule);
  	if(scaling == 1)                 
    	auto_scale(lp);
	if(scaling == 2)
    	scaleCR(lp);
   if(PRINT_SOLUTION >= 6)
    	print_scales(lp);
	result = solve(lp);				/* find solution */
  	if(print_timing)
    	print_cpu_times("solving");
	if(result == OPTIMAL) {
    	if (PRINT_SOLUTION >= 1)	print_objective(lp);
    	if (PRINT_SOLUTION >= 2)	print_solution(lp);
    	if (PRINT_SOLUTION >= 3)	print_constraints(lp);
    	if (PRINT_SOLUTION >= 4)	print_duals(lp);
      if(tracing)
      	fprintf(stderr, "Branch & Bound depth: %d\nNodes processed: %d\n"
         	"Simplex pivots: %d\n", get_max_level(lp),
            get_total_nodes(lp), get_total_iter(lp));
  	}
  	else if(result == INFEASIBLE)
    	{ if (PRINT_SOLUTION >= 1)	printf("This problem is infeasible\n"); }
  	else if(result == UNBOUNDED)
    	{ if (PRINT_SOLUTION >= 1)	printf("This problem is unbounded\n"); }
  	else if(result == FAILURE)
    	{ if (PRINT_SOLUTION >= 1)	printf("lp_solve failed\n"); }
  	delete_lp(lp);
  	return(result);
}

