/* xls2mps: convert lists of metabolites and reactions to .mps file */
/* also checks reactions for chemical balance (needs formulas in the first list) */
/* L.Rubanov, 2006 */
/* 30.04.06  1.4  Constant values changed. Precision increased of output */
/* 27.03.06  1.3  Chemical checking can be disabled (-C switch)
                  Console log output. Calculate balanced reactions. */
/* 30.09.04  1.2	Automatic N instead of E for 1st metabolite
                  Shortening metabolite and reaction names (8 & 8)
                  ...(e) -> ...e in extracellular metabolite names
                  Detail info about imbalance in formula
                  Remove bounds for unidirecional reactions
                  Insert delimiter lines in mps file */
/* 16.09.04  1.1  Added atoms: Na, Se
                  More strict checking of chemical formulas */
/* 07.09.04	 1.0	Base version */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#define	VERSION	"1.4"
#define	MAXABBR	1000			/* max number of metabolites */
#define	ABBRLEN	9				/* max length of metabolite name +1 */
#define	FORMLEN	64				/* max length of metabolite formula */
#define	MAXREACT	2000			/* max number of reactions */
#define	NAMELEN	8				/* max length of reaction name */
#define	REACTLEN	1024			/* max length of reaction notation */
#define	MAXEXCH	500			/* max number of exchange fluxes */

typedef struct {
	char	abbr[ABBRLEN];
   char	form[FORMLEN];
   int	C, O, H, N, P, S, Co, Fe, K, X, R, Na, Se;		/* atoms */
}	META;

typedef struct {
   int	revers;
	char	name[NAMELEN+1];
   char	form[REACTLEN];
}	REACTION;

typedef struct {
	char	name[NAMELEN+1];
   double lower, upper;
}	EXCHANGE;

int main( int argc, char** argv ) {
	META			*meta;
   REACTION	*react;
   EXCHANGE	*exch;
	FILE			*afile, *rfile, *efile, *mps;
   int	metanum, reactnum, exchnum, i, j, k, l, *pint;
   int	na=1, check=1, balanceok=0, revers=0, tlines=0;
   char	buf[2048], tmp[2048], ext, *ptr, subst[2*ABBRLEN];
   double	C, O, H, N, P, S, Co, Fe, K, X, R, Na, Se, d, hand;
   float		lo, up;
   printf("\n(C) 2006, Laboratory of Mathematical Methods and Models in Bioinformatics\n"
          "Institute for Information Transmission Problems, Russian Academy of Sciences\n"
   		 "Creating MPS file from Excel spreadsheet (see doc) v.%s     L.Rubanov, 2006\n\n", VERSION);
	if( argc<4 ) {
   	printf(
      	"Usage: xls2mps [-c] abbr_file react_file exch_file mps_file\n\n"
         "-c:         do not check chemical balance of reactions (check by default)\n"
         "abbr_file:  text file, each line shall be <abbr [chemical_formula]>\n"
         "react_file: text file, each line shall be <name reaction>\n"
         "exch_file:  text file, each line shall be <name reaction lower upper>\n"
         "mps_file:   output text file in MPS format\n\n"
         "Thank you for using the program.\n\n" );
      return 0;
   }
   if( (argv[na][0]=='-' || argv[na][0]=='/') &&
   	 (argv[na][1]=='c' || argv[na][1]=='C') )  {
   	check=0; na++;
      printf( "Warning: Chemical balance of reactions will not be checked!\n\n" );
   }
   afile=fopen( argv[na++], "rt" );
   rfile=fopen( argv[na++], "rt" );
   efile=fopen( argv[na++], "rt" );
   mps=fopen( argv[na], "wt" );
   if( !afile || !rfile || !efile || !mps ) {
   	printf( "Can't read or create file(s) - check arguments\n" );
      return -1;
   }
   meta=(META*)calloc( MAXABBR, sizeof(META) );
   react=(REACTION*)calloc( MAXREACT, sizeof(REACTION) );
   exch=(EXCHANGE*)calloc( MAXEXCH, sizeof(EXCHANGE) );
   if( !meta || !react || !exch) {
   	printf( "Can't allocate structure(s)\n" );
      return -1;
   }
   fprintf( mps, "NAME          from_xls\nROWS\n" ); tlines+=2;

   for( metanum=0; metanum<MAXABBR; metanum++ ) {	/* read abbr_file */
   	fgets( buf, sizeof( buf ), afile );
      if( feof( afile ) ) break;
      l=strlen( buf );
      if( l<=1 )
      	{ metanum--; continue; }
      if( buf[l-1]=='\n' || buf[l-1]=='\r' )	{ buf[l-1]=0; l--; }
      if( buf[l-1]=='\n' || buf[l-1]=='\r' )	  buf[l-1]=0; 
      												/* parse metabolite abbreviation */
      sscanf( buf, "%s", tmp );
      if( (l=strlen( tmp ))>=ABBRLEN )
      	printf( "Abbreviation too long in line %d - truncated\n", metanum+1 );
      strncpy( meta[metanum].abbr, tmp, ABBRLEN );
      meta[metanum].abbr[ABBRLEN-1]=0;
      if( check ) { 								/* parse metabolite formula */
         meta[metanum].C = meta[metanum].O = meta[metanum].H = meta[metanum].N = 0;
         meta[metanum].P = meta[metanum].S = meta[metanum].Co = 0;
         meta[metanum].Fe = meta[metanum].K = meta[metanum].X = meta[metanum].R = 0;
         sscanf( buf, "%*s%s", tmp );
         if( (l=strlen( tmp ))>=FORMLEN )
            printf( "Formula too long in line %d - truncated\n", metanum+1 );
         strncpy( meta[metanum].form, tmp, FORMLEN );
         meta[metanum].form[FORMLEN-1]=0;
         for( i=0; tmp[i]!=0; i++ ) {
            if( isspace( tmp[i] ) ) continue;
            if( isupper( tmp[i] ) ) {
               switch( tmp[i] ) {
                  case 'C':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].C;
                     else if( tmp[i+1]=='o' )
                        { pint=&meta[metanum].Co; i++; }
                     else pint=NULL;
                     break;
                  case 'O':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].O;
                     else pint=NULL;
                     break;
                  case 'H':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].H;
                     else pint=NULL;
                     break;
                  case 'N':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].N;
                     else if( tmp[i+1]=='a' )
                        { pint=&meta[metanum].Na; i++; }
                     else pint=NULL;
                     break;
                  case 'P':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].P;
                     else pint=NULL;
                     break;
                  case 'S':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].S;
                     else if( tmp[i+1]=='e' )
                        { pint=&meta[metanum].Se; i++; }
                     else pint=NULL;
                     break;
                  case 'F':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=NULL;
                     else if( tmp[i+1]=='e' )
                        { pint=&meta[metanum].Fe; i++; }
                     else pint=NULL;
                     break;
                  case 'K':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].K;
                     else pint=NULL;
                     break;
                  case 'X':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].X;
                     else pint=NULL;
                     break;
                  case 'R':
                     if( tmp[i+1]==0 || isdigit( tmp[i+1] ) || isupper( tmp[i+1] ) )
                        pint=&meta[metanum].R;
                     else pint=NULL;
                     break;
                  default:
                     pint=NULL;
               }
               if( !pint ) {
                  printf( "Unknown atom %c in line %d\n", tmp[i], metanum+1 );
                  break;
               }
               else if( *pint ) {
                  printf( "Repeated atom %c... in line %d\n", tmp[i], metanum+1 );
                  break;
               }
               if( !isdigit( tmp[i+1] ) )
                  { *pint=1; continue; }
               for( ++i; tmp[i]!=0; i++ ) {
                  if( !isdigit( tmp[i] ) ) break;
                  *pint = *pint*10+(int)(tmp[i]-'0');
               }
               --i;
            }
         }
      }
      fprintf( mps, " %c  %s\n", metanum?'E':'N', meta[metanum].abbr );
      tlines++;
   }
   if( metanum>=MAXABBR ) {
   	printf( "Too many metabolites - remainder ignored\n" );
   }
   printf( "%d metabolites accepted from file '%s'\n", metanum, argv[na-3] );
   fprintf( mps, "********************************************************\n" );
   fprintf( mps, "COLUMNS\n" );
   tlines+=2;
   													/* read react_file */
   for( reactnum=0; reactnum<MAXABBR; reactnum++ ) {
   	fgets( buf, sizeof( buf ), rfile );
      if( feof( rfile ) ) break;
      l=strlen( buf );
      if( l<=1 )
      	{ reactnum--; continue; }
      if( buf[l-1]=='\n' || buf[l-1]=='\r' )	{ buf[l-1]=0; l--; }
      if( buf[l-1]=='\n' || buf[l-1]=='\r' )	  buf[l-1]=0;
      												/* parse reaction name */
      sscanf( buf, "%s", tmp );
      if( (l=strlen( tmp ))>NAMELEN )
      	printf( "Reaction name too long in line %d - truncated\n", reactnum+1 );
      strncpy( react[reactnum].name, tmp, NAMELEN );
      react[reactnum].name[NAMELEN]=0;
      												/* parse&check reaction formula */
      C=O=H=N=P=S=Co=Fe=K=X=R=Na=Se=0.0;
      ext=0; hand=-1;
      strcpy( tmp, buf+l );
      if( (l=strlen( tmp ))>=REACTLEN )
      	printf( "Reaction too long in line %d - truncated\n", reactnum+1 );
      for( i=0; tmp[i]!=0 && isspace( tmp[i] ); i++ );
      strncpy( react[reactnum].form, tmp+i, REACTLEN );
      react[reactnum].form[REACTLEN-1]=0;
      react[reactnum].revers=0;
      for( ; tmp[i]!=0; i++ ) {
      	if( isspace( tmp[i] ) ) continue;
         if( tmp[i]=='[' ) {
         	i++;
            if( (tmp[i]=='c' && tmp[i+1]==']') ||
            	 (tmp[i]=='e' && tmp[i+1]==']') )
            	{ ext=tmp[i]; i=i+2; }
            else {
            	printf( "Invalid reaction compartment in line %d\n", reactnum+1 );
               break;
            }
         }
         for( j=i; tmp[j]!=0; j++ ) {		/* monomial parsing */
         	if( isspace( tmp[j] ) || tmp[j]=='+' ) continue;
            if( strncmp( tmp+j, "-->", 3 )==0 )
            	{ hand=1; j+=2; continue; }
            if( strncmp( tmp+j, "<==>", 4 )==0 )
            	{ hand=1; j+=3; react[reactnum].revers=1; continue; }
            if( tmp[j]=='(' ) {
            	d=atof( tmp+j+1 );				/* coefficient */
               ptr=strchr( tmp+j+1, ')' );
               if( !ptr ) {
               	printf( "No closing ) in line %d\n", reactnum+1 );
                  break;
               }
               j=(int)(ptr-tmp)+1;
               for( ; tmp[j]!=0 && isspace( tmp[j] ); j++ );
            }
            else d=1;
            subst[0]=0;
            for( k=j; tmp[k]!=0; k++ ) {
            	if( tmp[k]=='[' ) {
               	if( tmp[k+1]=='c' && tmp[k+2]==']' ) {
                  	strncpy( subst, tmp+j, k-j );
                     subst[k-j]=0;
                  }
                  else if( tmp[k+1]=='e' && tmp[k+2]==']' ) {
                  	strncpy( subst, tmp+j, k-j );
                     subst[k-j]=0;
                     strcat( subst, "e" );
                  }
                  else
                  	printf( "Invalid [ modifier in line %d\n", reactnum+1 );
                  j=k+2;
                  break;
               }
               if( tmp[k]==' ' ) {
               	strncpy( subst, tmp+j, k-j );
                  subst[k-j]=0;
                  if( ext=='e' )  strcat( subst, "e" );
                  j=k;
                  break;
               }
            }
            if( tmp[k]==0 ) {
            	strncpy( subst, tmp+j, k-j );
               subst[k-j]=0;
               if( ext=='e' )  strcat( subst, "e" );
               j=k-1;
            }
            if( (l=strlen(subst))==0 || l>ABBRLEN ) {
            	printf( "No name or name too long in line %d\n", reactnum+1 );
               break;
            }
            for( k=0; k<metanum; k++ ) {		/* check name with table */
            	if( strcmp( meta[k].abbr, subst )==0 ) break;
            }
            if( k>=metanum ) {
            	printf( "Unknown name '%s' in line %d\n", subst, reactnum+1 );
               break;
            }
            if( check ) {
               C += hand*d*meta[k].C;	O += hand*d*meta[k].O;
               H += hand*d*meta[k].H;	N += hand*d*meta[k].N;
               P += hand*d*meta[k].P;	S += hand*d*meta[k].S;
               Co += hand*d*meta[k].Co; Fe += hand*d*meta[k].Fe;
               K += hand*d*meta[k].K;	X += hand*d*meta[k].X;
               Na += hand*d*meta[k].Na; Se += hand*d*meta[k].Se;
               R += hand*d*meta[k].R;
            }
            fprintf( mps, "    %-10s%s%*g\n",
            	react[reactnum].name, subst, 22-strlen( subst ), hand*d );
            tlines++;
         }
         if( tmp[j]!=0 ) break;
         if( check && (fabs(C)>0.001 || fabs(O)>0.001 || 		/* check balance */
             fabs(H)>0.001 || fabs(N)>0.001 ||	fabs(P)>0.001 || fabs(S)>0.001 ||
             fabs(Co)>0.001 || fabs(Fe)>0.001 || fabs(K)>0.001 || fabs(X)>0.001 ||
             fabs(R)>0.001 || fabs(Na)>0.001 || fabs(Se)>0.001) ) {
         	printf( "Inequal balance in line %d ( ", reactnum+1 );
            if( fabs(C)>0.001 ) printf( "C=%.2f ", C );
            if( fabs(O)>0.001 ) printf( "O=%.2f ", O );
            if( fabs(H)>0.001 ) printf( "H=%.2f ", H );
            if( fabs(N)>0.001 ) printf( "N=%.2f ", N );
            if( fabs(P)>0.001 ) printf( "P=%.2f ", P );
            if( fabs(S)>0.001 ) printf( "S=%.2f ", S );
            if( fabs(Co)>0.001 ) printf( "Co=%.2f ", Co );
            if( fabs(Fe)>0.001 ) printf( "Fe=%.2f ", Fe );
            if( fabs(K)>0.001 ) printf( "K=%.2f ", K );
            if( fabs(X)>0.001 ) printf( "X=%.2f ", X );
            if( fabs(R)>0.001 ) printf( "R=%.2f ", R );
            if( fabs(Na)>0.001 ) printf( "Na=%.2f ", Na );
            if( fabs(Se)>0.001 ) printf( "Se=%.2f ", Se );
            printf( ")\n" );
         }
         else balanceok++;
         i=j-1;
      }
   }
   if( reactnum>=MAXREACT ) {
   	printf( "Too many reactions - remainder ignored\n" );
   }
   printf( "%d reactions accepted from file '%s'", reactnum, argv[na-2] );
   if( check ) printf( ", %d have proper chemical balance\n", balanceok );
   else printf( "\n" );
   fprintf( mps, "********************************************************\n" );
   tlines++;
   													/* read exchange_file */
   for( exchnum=0; exchnum<MAXEXCH; exchnum++ ) {
   	fgets( buf, sizeof( buf ), efile );
      if( feof( efile ) ) break;
      l=strlen( buf );
      if( l<=1 )
      	{ exchnum--; continue; }
      if( buf[l-1]=='\n' || buf[l-1]=='\r' )	{ buf[l-1]=0; l--; }
      if( buf[l-1]=='\n' || buf[l-1]=='\r' )	  buf[l-1]=0;
      												/* parse exchange reaction name */
      if( strncmp( buf, "EX_", 3 )!=0 ) {
      	printf( "Invalid exchange reaction name in line %d\n", exchnum+1 );
         continue;
      }
      sscanf( buf+2, "%s", tmp );		/* use only _... */
      ptr=strchr( tmp, '(' );
      if( ptr==NULL ) {
      	printf( "Invalid exchange reaction name in line %d\n", exchnum+1 );
         continue;
      }
      *ptr=0;
      if( (l=strlen( tmp ))>NAMELEN )
      	printf( "Exchange reaction name too long in line %d - truncated\n",
         	exchnum+1 );
      tmp[0]='X';					/* substitute _ with X */
      strncpy( exch[exchnum].name, tmp, NAMELEN );
      exch[exchnum].name[NAMELEN]=0;
      exch[exchnum].lower=-1e30;
      exch[exchnum].upper=1e30;
      												/* parse&check reaction formula */
      strcpy( tmp, buf+l+5 );
      for( i=0; tmp[i]!=0; i++ ) {
      	if( isspace( tmp[i] ) ) continue;
         if( strncmp( tmp+i, "[e]", 3 )!=0 )
         	printf( "No [e] in exchange reaction in line %d\n", exchnum+1 );
         sscanf( tmp+i+3, "%s", subst );
         if( strcmp( subst, exch[exchnum].name+1 )!=0 )
         	printf( "Inconsistent exchange names in line %d\n", exchnum+1 );
         ptr=strchr( tmp+i, '>' );
         if( ptr==0 )
         	printf( "No exchange symbol in line %d\n", exchnum+1 );
         if( strncmp( ptr-3, "<==>", 4 )!=0 )
         	printf( "Invalid exchange symbol in line %d\n", exchnum+1 );
         if( sscanf( ptr+1, "%e%e", &lo, &up )<2 )
         	printf( "Invalid exchange bound(s) in line %d\n", exchnum+1 );
         exch[exchnum].lower=lo;
         exch[exchnum].upper=up;
         break;
      }
      strcpy( tmp, exch[exchnum].name );
      fprintf( mps, "    %-10s%se%*g\n", tmp, tmp+1, 21-strlen( tmp+1 ), -1.0 );
      tlines++;
   }
   fprintf( mps, "********************************************************\n" );
   tlines++;
   fprintf( mps, "RHS\nBOUNDS\n" ); tlines+=2;
   for( k=0; k<exchnum; k++ ) {				/* exchange bounds */
   	strcpy( tmp, exch[k].name );
   	fprintf( mps, " LO BND       %s%*g\n", tmp,
      	22-strlen( tmp ), exch[k].lower );
   	fprintf( mps, " UP BND       %s%*g\n", tmp,
      	22-strlen( tmp ), exch[k].upper );
      tlines+=2;
   }
   printf( "%d exchange reactions with bounds accepted from file '%s'\n", exchnum, argv[na-1] );
   fprintf( mps, "********************************************************\n" );
   tlines++;
   for( k=0; k<reactnum; k++ ) {				/* other reactions bounds */
   	strcpy( tmp, react[k].name );
   	if( react[k].revers ) {
   		fprintf( mps, " LO BND       %s%*g\n", tmp, 22-strlen( tmp ), -1e30 );
   		fprintf( mps, " UP BND       %s%*g\n", tmp, 22-strlen( tmp ), +1e30 );
         revers++; tlines+=2;
      }
   }
   printf( "%d bound pairs added for reversible reactions\n", revers );
   fprintf( mps, "ENDATA\n" );
   tlines++;
   printf( "Total %d lines written into file '%s'\n\n", tlines, argv[na] );
   printf( "Thank you for using the program.\n\n" );
   if( mps ) fclose( mps );
   if( afile ) fclose( afile );
   if( rfile ) fclose( rfile );
   return 0;
}

